package jp.deadend.noname.mt

import android.content.Context
import android.graphics.drawable.Drawable
import android.widget.TextView
import android.os.Handler
import android.os.SystemClock
import android.graphics.Canvas
import android.os.Looper
import android.view.MotionEvent
import androidx.core.content.res.ResourcesCompat
import kotlin.math.abs

class FastScroller(context: Context, textView: TextView) {
    companion object {
        private const val ALPHA_MAX = 208
        private const val FADE_DURATION: Long = 500
        private const val MIN_PAGES = 3
        private const val STATE_NONE     = 0
        private const val STATE_VISIBLE  = 1
        private const val STATE_DRAGGING = 2
        private const val STATE_EXIT     = 3
    }

    private val mTextView = textView
    private var mScrollOrigin = 0
    private var mTotalHeight = -1
    private var isLongText = false

    private val mHandler = Handler(Looper.getMainLooper())
    private val mScrollFade = ScrollFade()

    private val mThumbDrawable: Drawable
    private val mThumbH: Int
    private val mThumbW: Int
    private var mThumbY: Int = 0

    private var mLastEventTime: Long = 0

    private var mState: Int

    init {
        val res = context.resources
        mThumbDrawable = requireNotNull(
            ResourcesCompat.getDrawable(res, R.drawable.fastscroller_thumb, null)
        ) { "BUG: Drawable for the fast scroller is null" }
        mThumbW = res.getDimensionPixelSize(R.dimen.fastscroll_thumb_width)
        mThumbH = res.getDimensionPixelSize(R.dimen.fastscroll_thumb_height)

        mState = STATE_NONE
    }

    fun setState(state: Int) {
        when (state) {
            STATE_NONE -> {
                mHandler.removeCallbacks(mScrollFade)
                mTextView.invalidate()
            }
            STATE_VISIBLE -> {
                if (mState != STATE_VISIBLE) { resetThumbPos() } // Optimization
                mHandler.removeCallbacks(mScrollFade)
            }
            STATE_DRAGGING -> mHandler.removeCallbacks(mScrollFade)
            STATE_EXIT -> mTextView.invalidate()
        }
        mState = state
    }

    private fun resetThumbPos() {
        val viewWidth = mTextView.width
        // Bounds are always top right. Y coordinate get's translated during draw
        mThumbDrawable.setBounds(viewWidth - mThumbW, 0, viewWidth, mThumbH)
        mThumbDrawable.alpha = ALPHA_MAX
    }

    fun isVisible(): Boolean = (mState != STATE_NONE)

    fun draw(canvas: Canvas) {
        if (mState == STATE_NONE) return

        val x = mTextView.scrollX
        val y = mThumbY + mTextView.scrollY

        var alpha = -1
        if (mState == STATE_EXIT) {
            alpha = mScrollFade.alpha
            if (alpha < ALPHA_MAX / 2) { mThumbDrawable.alpha = alpha * 2 }
        }

        canvas.translate(x.toFloat(), y.toFloat())
        mThumbDrawable.draw(canvas)
        canvas.translate(-x.toFloat(), -y.toFloat())

        if (alpha == 0) {
            setState(STATE_NONE)
        } else {
            mTextView.invalidate()
        }
    }

    fun onScroll(position: Int, visibleHeight: Int, totalHeight: Int) {
        // Are there enough pages to require fast scroll? Recompute only if total height changes
        if (mTotalHeight != totalHeight && visibleHeight > 0) {
            mTotalHeight = totalHeight
            isLongText = mTotalHeight / visibleHeight >= MIN_PAGES
        }
        if (!isLongText) {
            if (mState != STATE_NONE) { setState(STATE_NONE) }
            return
        }

        if (totalHeight - visibleHeight > 0 && mState != STATE_DRAGGING) {
            mThumbY = ((mTextView.height - mThumbH) * position / (totalHeight - visibleHeight))
        }

        if (position == mScrollOrigin) return
        mScrollOrigin = position
    }

    private fun scrollTo(position: Float) {
        (mTextView as? MyEditText)?.moveToLine((position * mTextView.lineCount).toInt())
    }

    private fun cancelFling() {
        val cancelFling = MotionEvent.obtain(0, 0, MotionEvent.ACTION_CANCEL, 0f, 0f, 0)
        mTextView.onTouchEvent(cancelFling)
        cancelFling.recycle()
    }

    private fun isPointInside(x: Float, y: Float) =
        x > mTextView.width - mThumbW && y >= mThumbY && y <= mThumbY + mThumbH

    fun onInterceptTouchEvent(ev: MotionEvent): Boolean {
        if (mState > STATE_NONE && ev.action == MotionEvent.ACTION_DOWN && isPointInside(ev.x, ev.y)) {
            setState(STATE_DRAGGING)
            return true
        } else if (isLongText && mState == STATE_NONE && ev.action == MotionEvent.ACTION_MOVE) {
            setState(STATE_VISIBLE)
            mHandler.postDelayed(mScrollFade, 1500)
        }
        return false
    }

    fun onTouchEvent(me: MotionEvent): Boolean {
        if (mState == STATE_NONE) { return false }

        when (me.action) {
            MotionEvent.ACTION_DOWN -> {
                if (isPointInside(me.x, me.y)) {
                    setState(STATE_DRAGGING)
                    cancelFling()
                    return true
                }
            }
            MotionEvent.ACTION_UP -> {
                if (mState == STATE_DRAGGING) {
                    setState(STATE_VISIBLE)
                    val handler = mHandler
                    handler.removeCallbacks(mScrollFade)
                    handler.postDelayed(mScrollFade, 1000)
                    return true
                }
            }
            MotionEvent.ACTION_MOVE -> {
                if (mState == STATE_DRAGGING) {
                    val now = System.currentTimeMillis()
                    val diff = now - mLastEventTime
                    if (diff > 30) {
                        mLastEventTime = now

                        val viewHeight = mTextView.height
                        // Jitter
                        val ypos = me.y.toInt() - mThumbH / 2
                        val newThumbY = when {
                            ypos < 0 -> 0
                            ypos + mThumbH > viewHeight -> viewHeight - mThumbH
                            else -> ypos
                        }
                        if (abs(mThumbY - newThumbY) < 2) { return true }
                        mThumbY = newThumbY

                        scrollTo(mThumbY.toFloat() / (viewHeight - mThumbH))
                    }
                    return true
                }
            }
        }

        return false
    }

    inner class ScrollFade : Runnable {
        private var mStartTime: Long = 0
        private var mFadeDuration: Long = 0

        val alpha: Int
            get() {
                if (mState != STATE_EXIT) { return ALPHA_MAX }

                val now = SystemClock.uptimeMillis()
                return if (now > mStartTime + mFadeDuration) {
                    0
                } else {
                    (ALPHA_MAX - (now - mStartTime) * ALPHA_MAX / mFadeDuration).toInt()
                }
            }

        override fun run() {
            if (mState != STATE_EXIT) {
                mFadeDuration = FADE_DURATION
                mStartTime = SystemClock.uptimeMillis()
                setState(STATE_EXIT)
                return
            }

            if (alpha > 0) {
                mTextView.invalidate()
            } else {
                setState(STATE_NONE)
            }
        }
    }
}