package jp.deadend.noname.mt

import android.app.Dialog
import android.content.*
import android.graphics.Typeface
import android.net.Uri
import android.os.Bundle
import android.view.Menu
import android.view.MenuItem
import android.view.View
import androidx.activity.OnBackPressedCallback
import androidx.activity.result.contract.ActivityResultContracts
import androidx.appcompat.app.AlertDialog
import androidx.appcompat.app.AppCompatActivity
import androidx.appcompat.widget.SearchView
import androidx.appcompat.widget.Toolbar
import androidx.core.view.WindowInsetsCompat
import androidx.core.view.WindowInsetsControllerCompat
import androidx.fragment.app.DialogFragment
import androidx.preference.PreferenceManager
import com.google.android.material.snackbar.Snackbar
import jp.deadend.noname.mt.databinding.ActivityMainBinding
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.MainScope
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import java.io.BufferedReader
import java.io.File
import java.io.FileOutputStream
import java.io.InputStreamReader
import java.util.*

class MainActivity : AppCompatActivity(), MyEditText.UndoStateCallback, LinearLayoutKeyDispatcher.Callback {
    private var mCurrentUri: Uri? = null
    private var mShortcutKeys: Map<Int, EditorCommand>? = null
    private lateinit var binding: ActivityMainBinding
    private var lastSearchTime = 0L

    private val settingsActivity =
        registerForActivityResult(ActivityResultContracts.StartActivityForResult()) { loadPrefs() }
    private val openFileActivity =
        registerForActivityResult(ActivityResultContracts.OpenDocument()) { uri ->
            if (uri != null) {
                contentResolver.takePersistableUriPermission(
                    uri,
                    Intent.FLAG_GRANT_READ_URI_PERMISSION or Intent.FLAG_GRANT_WRITE_URI_PERMISSION
                )
                loadFromUri(uri, 0)
            }
        }
    private val saveAsActivity =
        registerForActivityResult(ActivityResultContracts.CreateDocument("text/plain")) { uri ->
            if (uri != null) {
                contentResolver.takePersistableUriPermission(
                    uri,
                    Intent.FLAG_GRANT_READ_URI_PERMISSION or Intent.FLAG_GRANT_WRITE_URI_PERMISSION
                )
                saveToUri(uri)
            }
        }

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        binding = ActivityMainBinding.inflate(layoutInflater)
        setContentView(binding.root)
        supportActionBar?.setTitle(R.string.label_untitled)
//        binding.toolbarSecond.inflateMenu(R.menu.menu_second)
//        binding.toolbarSecond.setOnMenuItemClickListener(object : Toolbar.OnMenuItemClickListener {
//            override fun onMenuItemClick(item: MenuItem): Boolean {
//                when (item.itemId) {
//                    R.id.menu_second_cut -> performCommand(EditorCommand.CUT)
//                    R.id.menu_second_copy -> performCommand(EditorCommand.COPY)
//                    R.id.menu_second_paste -> performCommand(EditorCommand.PASTE)
//                    R.id.menu_second_first -> performCommand(EditorCommand.GOTO_LINE_BEGINNING)
//                    R.id.menu_second_last -> performCommand(EditorCommand.GOTO_LINE_END)
//                }
//                return false
//            }
//        })
        binding.toolbarFind.inflateMenu(R.menu.menu_find)
        binding.toolbarFind.setOnMenuItemClickListener(object : Toolbar.OnMenuItemClickListener {
            override fun onMenuItemClick(item: MenuItem): Boolean {
                when (item.itemId) {
                    R.id.menu_find_cancel -> {
                        binding.toolbarFind.visibility = View.GONE
                        binding.editTextMain.clearHighlight()
                        return true
                    }
                    R.id.menu_find_prev -> performCommand(EditorCommand.FIND_PREV)
                    R.id.menu_find_next -> performCommand(EditorCommand.FIND_NEXT)
                }
                return false
            }
        })
        val searchView = binding.toolbarFind.findViewById<SearchView>(R.id.menu_find_searchview)
        searchView.setIconifiedByDefault(false)
        searchView.setOnQueryTextListener(object : SearchView.OnQueryTextListener {
            override fun onQueryTextSubmit(query: String): Boolean {
                if (query.isNotEmpty()) {
                    val time = Calendar.getInstance().timeInMillis
                    if (time > lastSearchTime + 500) {
                        performCommand(EditorCommand.FIND_NEXT)
                        lastSearchTime = time // workaround not to run twice
                    }
                    return true
                }
                return false
            }

            override fun onQueryTextChange(newText: String): Boolean {
                if (newText.isNotEmpty()) {
                    binding.editTextMain.let {
                        it.findAndHighlight(newText)
                        if (it.selectionStart > 0) { it.setSelection(it.selectionStart - 1) }
                        it.gotoNextOrPrevHighlight(true)
                    }
                    return true
                }
                return false
            }
        })

        binding.editTextMain.setCallback(this)
        binding.layoutMain.setCallback(this)

//        val data = intent.data
//        data?.let {
//            if (data.scheme.equals("file") && data.path != null) {
//                val file = File(data.path!!)
//                mCurrentDir = file.parent
//                mCurrentFile = file.name
//                loadFile(file)
//            }
//        }

        if (prefs.lastUri != null) {
            loadFromUri(Uri.parse(prefs.lastUri), prefs.lastOffset)
        }

        loadPrefs()

        onBackPressedDispatcher.addCallback(this, object : OnBackPressedCallback(true) {
            override fun handleOnBackPressed() {
                confirmSave(::finish)
            }
        })
    }

    override fun onUndoStateChanged() = invalidateOptionsMenu()

    override fun onCreateOptionsMenu(menu: Menu): Boolean {
        menuInflater.inflate(R.menu.menu_main, menu)
        menu.findItem(R.id.menu_main_show_line_numbers).isChecked = prefs.showLineNumbers
        menu.findItem(R.id.menu_main_fullscreen).isChecked = prefs.fullScreen

        return true
    }

    override fun onPrepareOptionsMenu(menu: Menu): Boolean {
        for (i in 0 until menu.size()) {
            val item = menu.getItem(i)
            when (item.itemId) {
                R.id.menu_main_undo -> item.isEnabled = binding.editTextMain.isUndoable()
                R.id.menu_main_redo -> item.isEnabled = binding.editTextMain.isRedoable()
                R.id.menu_main_save -> item.isEnabled = binding.editTextMain.isUndoable()
            }
        }

        return super.onPrepareOptionsMenu(menu)
    }

    override fun onOptionsItemSelected(item: MenuItem) : Boolean {
        when (item.itemId) {
            R.id.menu_main_undo -> performCommand(EditorCommand.UNDO)
            R.id.menu_main_redo -> performCommand(EditorCommand.REDO)
            R.id.menu_main_cut -> performCommand(EditorCommand.CUT)
            R.id.menu_main_copy -> performCommand(EditorCommand.COPY)
            R.id.menu_main_paste -> performCommand(EditorCommand.PASTE)
            R.id.menu_main_cut_line -> performCommand(EditorCommand.CUT_LINE)
            R.id.menu_main_open -> performCommand(EditorCommand.OPEN)
            R.id.menu_main_save, R.id.menu_main_save2 -> performCommand(EditorCommand.SAVE)
            R.id.menu_main_save_as -> performCommand(EditorCommand.SAVE_AS)
            R.id.menu_main_close -> performCommand(EditorCommand.CLOSE)
            R.id.menu_main_settings -> settingsActivity.launch(
                Intent(this, SettingsActivity::class.java)
            )
            R.id.menu_main_top -> performCommand(EditorCommand.GOTO_TOP)
            R.id.menu_main_bottom -> performCommand(EditorCommand.GOTO_BOTTOM)
            R.id.menu_main_line_beginning -> performCommand(EditorCommand.GOTO_LINE_BEGINNING)
            R.id.menu_main_line_end -> performCommand(EditorCommand.GOTO_LINE_END)
            R.id.menu_main_find -> performCommand(EditorCommand.FIND)
            R.id.menu_find_prev -> performCommand(EditorCommand.FIND_PREV)
            R.id.menu_find_next -> performCommand(EditorCommand.FIND_NEXT)
            R.id.menu_main_show_line_numbers -> {
                item.isChecked = !item.isChecked
                binding.editTextMain.showLineNumbers = item.isChecked
            }
            R.id.menu_main_fullscreen -> {
                item.isChecked = !item.isChecked
                setFullScreen(item.isChecked)
                prefs.fullScreen = item.isChecked
            }
            R.id.menu_main_read_only  -> {
                item.isChecked = !item.isChecked
                binding.editTextMain.isReadonly = item.isChecked
                if (item.isChecked) {
                    supportActionBar?.title?.let { supportActionBar?.title = "(RO) $it" }
                } else {
                    supportActionBar?.title?.let { supportActionBar?.title = it.subSequence(5, it.length) }
                }
            }
            R.id.menu_main_count -> {
                val chars = binding.editTextMain.length()
                val lines = binding.editTextMain.countLogicalLines()
                SimpleMessageDialogFragment.newInstance(
                    getString(R.string.message_count, chars, lines)
                ).show(supportFragmentManager, "dialog")
            }
        }

        return super.onOptionsItemSelected(item)
    }

    override fun onPause() {
        if (mCurrentUri != null) {
            prefs.lastUri = mCurrentUri.toString()
            prefs.lastOffset = binding.editTextMain.selectionStart
        } else {
            PreferenceManager.getDefaultSharedPreferences(this)
                .edit()
                .remove(getString(R.string.prefkey_last_uri))
                .remove(getString(R.string.prefkey_last_offset))
                .apply()
        }
        prefs.showLineNumbers = binding.editTextMain.showLineNumbers

        super.onPause()
    }

    private fun setFullScreen(fullScreen: Boolean) {
        if (fullScreen) {
            WindowInsetsControllerCompat(window, binding.root).also { controller ->
                controller.hide(
                    WindowInsetsCompat.Type.statusBars()
                        .or(WindowInsetsCompat.Type.navigationBars())
                )

                controller.systemBarsBehavior = WindowInsetsControllerCompat.BEHAVIOR_SHOW_TRANSIENT_BARS_BY_SWIPE
            }
        } else {
            WindowInsetsControllerCompat(window, binding.root).also { controller ->
                controller.show(
                    WindowInsetsCompat.Type.statusBars()
                        .or(WindowInsetsCompat.Type.navigationBars())
                )
            }
        }
    }

    private fun loadPrefs() {
        binding.editTextMain.let {
            it.textSize = prefs.fontSize.toFloat()
            it.setLineSpacing(0f, prefs.lineSpacing / 10f)

            if (prefs.fullScreen) { setFullScreen(true) }

            if (prefs.useCustomFont) {
                val fontFile = File(filesDir, prefs.customFontName)
                it.typeface = if (fontFile.canRead()) Typeface.createFromFile(fontFile) else null
            } else {
                it.typeface = null
            }
            it.setBackgroundColor(prefs.colorBackground)
            it.setTextColor(prefs.colorText)
            it.setCurrentLineHighlightColor(prefs.colorCurrentLine)
            it.showLineNumbers = prefs.showLineNumbers

            it.recalcForDraw()
        }
        binding.layoutMain.setBackgroundColor(prefs.colorBackground)

        mShortcutKeys = prefs.getShortcuts()
    }

    override fun commandShortcut(key: Int) : Boolean {
        val shortcuts = mShortcutKeys ?: return false
        val command = shortcuts[key] ?: return false
        performCommand(command)
        return true
    }

    private fun performCommand(command: EditorCommand) {
        val editText = binding.editTextMain
        when (command) {
            EditorCommand.UNDO -> editText.undo()
            EditorCommand.REDO -> editText.redo()
            EditorCommand.GOTO_TOP -> editText.setSelection(0)
            EditorCommand.GOTO_BOTTOM -> editText.gotoBottom()
            EditorCommand.GOTO_LINE_BEGINNING -> editText.gotoLineBeginning()
            EditorCommand.GOTO_LINE_END -> editText.gotoLineEnd()
            EditorCommand.CUT_LINE -> binding.editTextMain.cutCurrentLine()?.let { setClip(it) }
            EditorCommand.CUT -> binding.editTextMain.cutSelection()?.let { setClip(it) }
            EditorCommand.COPY -> binding.editTextMain.copySelection()?.let { setClip(it) }
            EditorCommand.PASTE -> {
                (getSystemService(Context.CLIPBOARD_SERVICE) as ClipboardManager)
                    .primaryClip?.getItemAt(0)?.coerceToText(this)
                    ?.let { binding.editTextMain.pasteText(it) }
            }
            EditorCommand.OPEN -> confirmSave { openFileActivity.launch(arrayOf("text/*")) }
            EditorCommand.SAVE -> {
                val uri = mCurrentUri
                if (uri == null) {
                    performCommand(EditorCommand.SAVE_AS)
                } else {
                    saveToUri(uri)
                }
            }
            EditorCommand.SAVE_AS -> {
                editText.text?.let {
                    val uri = mCurrentUri
                    val offset = it.indexOf('\n')

                    val title = when {
                        uri != null     -> (application as MainApplication).getFileName(uri)
                        offset > 0      -> it.subSequence(0, offset).toString() + ".txt"
                        it.isNotEmpty() -> "$it.txt"
                        else            -> getString(R.string.filename_untitled)
                    }

                    saveAsActivity.launch(title)
                }
            }
            EditorCommand.CLOSE -> {
                confirmSave {
                    binding.editTextMain.loadNewText("")
                    mCurrentUri = null
                    supportActionBar?.setTitle(R.string.label_untitled)
                }
            }
            EditorCommand.FIND -> {
                if (binding.toolbarFind.visibility == View.GONE) {
                    binding.toolbarFind.visibility = View.VISIBLE
                    binding.toolbarFind.findViewById<SearchView>(R.id.menu_find_searchview).requestFocus()
                } else {
                    binding.toolbarFind.visibility = View.GONE
                    editText.clearHighlight()
                }
            }
            EditorCommand.FIND_NEXT -> {
                if (editText.isHighlighted()) { editText.gotoNextOrPrevHighlight(true) }
            }
            EditorCommand.FIND_PREV -> {
                if (editText.isHighlighted()) { editText.gotoNextOrPrevHighlight(false) }
            }
        }
    }

    private fun setClip(s: CharSequence) {
        (getSystemService(Context.CLIPBOARD_SERVICE) as ClipboardManager)
            .setPrimaryClip(ClipData.newPlainText("clip", s))
    }

    private fun confirmSave(f: () -> Unit) {
        if (!binding.editTextMain.isUndoable()) {
            f()
        } else {
            val dialog = ConfirmSaveDialogFragment()
            dialog.setListener(
                object : ConfirmSaveDialogFragment.Listener {
                    override fun onPositiveClick() = performCommand(EditorCommand.SAVE)
                    override fun onNegativeClick() = f()
                }
            )
            dialog.show(supportFragmentManager, "dialog")
        }
    }

    private suspend fun readTextFromUri(uri: Uri): String? {
        return withContext(Dispatchers.IO) {
            try {
                val builder = StringBuilder()
                contentResolver.openInputStream(uri)?.use { inputStream ->
                    BufferedReader(InputStreamReader(inputStream)).use { reader ->
                        var line: String? = reader.readLine()
                        while (line != null) {
                            builder.append(line)
                            builder.append('\n')
                            line = reader.readLine()
                        }
                    }
                }
//                contentResolver.openInputStream(uri)
//                    ?.bufferedReader(charset("UTF-8"))
//                    ?.forEachLine { line ->
//                        builder.append(line)
//                        builder.append('\n')
//                    }
                builder.toString()
            } catch (e: Exception) {
                Trace.d("Exception: $e")
                null
            }
        }
    }

    private fun loadFromUri(uri: Uri, offset: Int) {
        MainScope().launch {
            binding.progressBar.visibility = View.VISIBLE
            val text = readTextFromUri(uri)
            if (text != null) {
                binding.editTextMain.loadNewText(text)
                mCurrentUri = uri

                supportActionBar?.title = (application as MainApplication).getFileName(uri) ?: ""
            } else {
                SimpleMessageDialogFragment.newInstance(
                    getString(R.string.error_open_failed, uri)
                ).show(supportFragmentManager, "dialog")
            }
            binding.progressBar.visibility = View.GONE
            val edit = binding.editTextMain
            val limit = edit.text?.length ?: 0
            if (offset in 0 .. limit) { edit.setSelection(offset) }
        }
    }

    private suspend fun writeTextToUri(uri: Uri) {
        withContext(Dispatchers.IO) {
            contentResolver.openFileDescriptor(uri, "wt")?.use {
                FileOutputStream(it.fileDescriptor).use { stream ->
                    stream.write(binding.editTextMain.text.toString().toByteArray())
                }
            }
//            contentResolver.openOutputStream(uri, "wt")
//                ?.bufferedWriter(charset("UTF-8"))?.use {
//                    it.write(binding.editTextMain.text.toString())
//                }
        }
    }

    private fun saveToUri(uri: Uri) {
        MainScope().launch {
            binding.progressBar.visibility = View.VISIBLE
            try {
                writeTextToUri(uri)
                binding.editTextMain.clearEditHistory()
                Snackbar.make(
                    binding.layoutMain,
                    getString(
                        R.string.message_saved,
                        (application as MainApplication).getFileName(uri)
                    ),
                    Snackbar.LENGTH_LONG
                ).show()

                mCurrentUri = uri
                supportActionBar?.title = (application as MainApplication).getFileName(uri) ?: ""
            } catch (e: Exception) {
                Trace.d("Exception: $e")
                SimpleMessageDialogFragment.newInstance(
                    getString(R.string.error_save_failed, uri)
                ).show(supportFragmentManager, "dialog")
            }
            binding.progressBar.visibility = View.GONE
        }
    }

    class ConfirmSaveDialogFragment : DialogFragment() {
        private var mListener: Listener? = null

        interface Listener {
            fun onPositiveClick()
            fun onNegativeClick()
        }

        fun setListener(listener: Listener) {
            this.mListener = listener
        }

        override fun onCreateDialog(savedInstanceState: Bundle?): Dialog {
            return AlertDialog.Builder(requireContext())
                .setMessage(R.string.message_confirm_save)
                .setCancelable(true)
                .setPositiveButton(R.string.label_do_save) { _, _ ->
                    mListener?.onPositiveClick()
                    dismiss()
                }
                .setNegativeButton(R.string.label_dont_save) { _, _ ->
                    mListener?.onNegativeClick()
                    dismiss()
                }
                .setNeutralButton(android.R.string.cancel) { _, _ ->
                    dismiss()
                }
                .create()
        }
    }
}