package jp.deadend.noname.mt

import android.content.Context
import android.graphics.*
import android.text.*
import android.text.style.BackgroundColorSpan
import android.util.AttributeSet
import android.util.SparseIntArray
import android.util.TypedValue

class MyEditText(context: Context, attrs: AttributeSet) : AppCompatEditTextWithFastScroller(context, attrs) {
    private data class EditHistoryItem(
        val position: Int,
        val before: CharSequence,
        var after: CharSequence
    )
    private val mEditHistory = ArrayDeque<EditHistoryItem>()
    private var mRedoIndex = 0
    private var isRecordingHistory = true
    private var mHighlightSpan: BackgroundColorSpan? = null

    interface UndoStateCallback { fun onUndoStateChanged() }
    private var mCallback: UndoStateCallback? = null
    fun setCallback(c: UndoStateCallback) { mCallback = c }

    private val mCurrentLineRect = Rect()
    var showLineNumbers = true
        set(value) {
            field = value
            val lpad = if (value) mNumPadding*mNumPaddingCount else paddingRight
            setPadding(lpad, paddingTop, paddingRight, paddingBottom)
        }
    private val mNumPaint = TextPaint(Paint.ANTI_ALIAS_FLAG).also {
        it.color = Color.DKGRAY
        it.typeface = Typeface.MONOSPACE
        it.isDither = false
    }
    private var mNumPadding = 0
    private var mNumPaddingCount = 0
    private val mCurrentLineHighlightPaint = Paint(Paint.ANTI_ALIAS_FLAG).also {
        it.color = Color.LTGRAY
    }
    private var mCurrentLineCount = 0
    private val mLineNumberPos = SparseIntArray()

    var isReadonly = false
        set(value) {
            field = value
            isFocusable = !value
            isFocusableInTouchMode = !value
            isClickable = !value
            isLongClickable = !value
            isCursorVisible = !value
        }

    init {
        addTextChangedListener(object : TextWatcher {
            override fun beforeTextChanged(s: CharSequence, start: Int, count: Int, after: Int) {
                if (isRecordingHistory) {
                    for (i in mRedoIndex until mEditHistory.size) { mEditHistory.removeLast() }
                    mEditHistory.add(
                        EditHistoryItem(start, s.subSequence(start, start + count), "")
                    )
                    mRedoIndex = mEditHistory.size
                }
                if (mEditHistory.size == 1) { mCallback?.onUndoStateChanged() }
            }
            override fun onTextChanged(s: CharSequence, start: Int, before: Int, count: Int) {
                if (isRecordingHistory) {
                    mEditHistory.last().after = s.subSequence(start, start + count)
                }
            }
            override fun afterTextChanged(s: Editable) { }
        })
    }

    fun isUndoable() = mEditHistory.isNotEmpty() && mRedoIndex > 0
    fun isRedoable() = mRedoIndex < mEditHistory.size
    fun clearEditHistory() {
        mEditHistory.clear()
        mRedoIndex = 0
        mCallback?.onUndoStateChanged()
    }

    fun undo() {
        if (!isUndoable()) { return }

        val item = mEditHistory[mRedoIndex - 1]
        isRecordingHistory = false
        text?.replace(item.position, item.position + item.after.length, item.before)
        isRecordingHistory = true
        mRedoIndex--

        mCallback?.onUndoStateChanged()
    }

    fun redo() {
        if (!isRedoable()) { return }

        val item = mEditHistory[mRedoIndex]
        isRecordingHistory = false
        text?.replace(item.position, item.position + item.before.length, item.after)
        isRecordingHistory = true
        mRedoIndex++

        mCallback?.onUndoStateChanged()
    }

    fun loadNewText(text: CharSequence) {
        super.setText(text)
        clearEditHistory()
    }

    fun moveToLine(line: Int) {
        setSelection(layout.getLineStart(line))
    }

    fun setCurrentLineHighlightColor(color: Int) {
        mCurrentLineHighlightPaint.color = color
        invalidate()
    }

//    private fun getCurrentLogicalLine(position: Int): Int {
//        text?.let {
//            var count = 0
//            for (i in position-1 downTo 0) {
//                if (it[i] == '\n') { count++ }
//            }
//            return count
//        }
//        return 0
//    }

    fun countLogicalLines(): Int {
        text?.let { t ->
            return t.count { it == '\n' } + 1
        }
        return 1
    }

    private fun getLogicalLineStart(position: Int): Int {
        text?.let {
            for (i in position-1 downTo 0) {
                if (it[i] == '\n') { return i+1 }
            }
        }
        return 0
    }
    private fun getLogicalLineEnd(position: Int): Int {
        text?.let {
            for (i in position until it.length) {
                if (it[i] == '\n') { return i }
            }
            return it.length
        }
        return 0
    }

    fun cutCurrentLine(): CharSequence? {
        text?.let {
            val start = getLogicalLineStart(selectionStart)
            val tmp = getLogicalLineEnd(selectionStart)
            val end = if (tmp == it.length) tmp else tmp+1
            val result = it.subSequence(start, end)
            it.delete(start, end)
            return result
        }
        return null
    }

    fun cutSelection(): CharSequence? {
        text?.let {
            val result = it.subSequence(selectionStart, selectionEnd)
            it.delete(selectionStart, selectionEnd)
            return result
        }
        return null
    }

    fun copySelection(): CharSequence? {
        text?.let { return it.subSequence(selectionStart, selectionEnd) }
        return null
    }

    fun pasteText(s: CharSequence) { text?.insert(selectionStart, s) }

    fun gotoBottom() = setSelection(text?.length ?: 0)
    fun gotoLineBeginning() = setSelection(getLogicalLineStart(selectionStart))
    fun gotoLineEnd() = setSelection(getLogicalLineEnd(selectionStart))

    fun clearHighlight() {
        val editable = text ?: return
        val spans = editable.getSpans(0, editable.length, BackgroundColorSpan::class.java)
        for (span in spans) { editable.removeSpan(span) }
        mHighlightSpan?.let { editable.removeSpan(it) }
    }

    fun isHighlighted(): Boolean {
        text?.let {
            return it.getSpans(0, it.length, BackgroundColorSpan::class.java).isNotEmpty()
        }
        return false
    }

    fun findAndHighlight(str: String): Boolean {
        clearHighlight()
        if (str.isEmpty()) { return false }

        val indices: MutableList<Int> = mutableListOf()
        var idx = 0
        val t = text ?: return false
        while (idx <= t.lastIndex) {
            idx = t.indexOf(str, idx+1, true)
            if (idx == -1) break
            indices.add(idx)
        }

        if (indices.isNotEmpty()) {
            for (i in indices) {
                t.setSpan(BackgroundColorSpan(Color.YELLOW), i, i + str.length, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
            }
            return true
        }

        return false

    }

    fun gotoNextOrPrevHighlight(forward: Boolean) {
        val editable = text ?: return
        val spans = editable.getSpans(0, editable.length, BackgroundColorSpan::class.java)

        val targetSpan = if (forward) {
            spans.find { editable.getSpanStart(it) > selectionStart }
        } else {
            spans.findLast { editable.getSpanEnd(it) < selectionStart }
        }
        targetSpan?.let {
            setSelection(editable.getSpanStart(it))
            mHighlightSpan?.let { span -> editable.removeSpan(span) }
            mHighlightSpan = BackgroundColorSpan(Color.RED)
            editable.setSpan(
                mHighlightSpan,
                editable.getSpanStart(it),
                editable.getSpanEnd(it),
                Spannable.SPAN_EXCLUSIVE_EXCLUSIVE
            )
        }
    }

    override fun setTextSize(size: Float) {
        mNumPaint.textSize = TypedValue.applyDimension(
            TypedValue.COMPLEX_UNIT_SP, size/2, resources.displayMetrics
//            TypedValue.COMPLEX_UNIT_SP, size, resources.displayMetrics
        )
        mNumPadding = mNumPaint.measureText("8").toInt()
        super.setTextSize(size)
    }

    private fun calcLinePos(lo: Layout, str: String) {
        mCurrentLineCount = lineCount

        mLineNumberPos.clear()
        var lineNo = 1
        for (i in 0 until lineCount) {
            val st = lo.getLineStart(i)
            if (i == 0 || str[st - 1] == '\n') {
                mLineNumberPos.append(i, lineNo)
                lineNo++
            }
        }

        val newCount = (lineNo-1).toString().length + 1
        if (newCount != mNumPaddingCount) {
            mNumPaddingCount = newCount
            if (showLineNumbers) {
                setPadding(mNumPadding * newCount, paddingTop, paddingRight, paddingBottom)
            }
        }
    }

    fun recalcForDraw() {
        text?.toString()?.let { str ->
            layout?.let { lo -> calcLinePos(lo, str) }
        }
    }

    override fun onDraw(canvas: Canvas) {
        text?.toString()?.let { str ->
            layout?.let { lo ->
                if (lineCount != mCurrentLineCount) { calcLinePos(lo, str) }

                if (!isReadonly) {
                    getLineBounds(
                        lo.getLineForOffset(getLogicalLineStart(selectionStart)),
                        mCurrentLineRect
                    )
                    val top = mCurrentLineRect.top
                    getLineBounds(
                        lo.getLineForOffset(getLogicalLineEnd(selectionStart)),
                        mCurrentLineRect
                    )
                    mCurrentLineRect.top = top
                    canvas.drawRect(mCurrentLineRect, mCurrentLineHighlightPaint)
                }

                if (str.isEmpty()) {
                    if (showLineNumbers) {
                        canvas.drawText("1", 0f, -paint.fontMetrics.top + paddingTop, mNumPaint)
                    }
                } else {
                    val first = lo.getLineForVertical(scrollY)
                    val last = lo.getLineForVertical(height + scrollY)
                    for (i in first .. last) {
                        val baseLine = lo.getLineBottom(i).toFloat()
                        if (showLineNumbers) {
                            if (mLineNumberPos[i] != 0) {
                                canvas.drawText(mLineNumberPos[i].toString(), 0f, baseLine, mNumPaint)
                            }
                        }
                        val st = lo.getLineStart(i)
                        val en = lo.getLineEnd(i)
                        if (st != en && str[en - 1] == '\n') {
                            canvas.drawText("⏎",
                                paint.measureText(str.substring(st until en)) + paddingLeft,
                                baseLine,
                                mNumPaint
                            )
                        }
                    }
                }
            }
        }

        super.onDraw(canvas)
    }
}